"""Functions for calculating the activity of water in glycerol-water solutions based on Zhang 2022.

Source
------
- Zhang, L., Grace, P. M. & Sun, D.-W. 
An accurate water activity model for glycerol solutions and its implementation on moisture sorption isotherm determination.
Drying Technology 40, 2404-2413 (2022).
"""


import numpy as np

def water_activity_glycerol(x, T):
    """
    Calculate the activity of water in a glycerol-water solution based on mole fraction and temperature (K).

    Wilson model from:
    Zhang, L., Grace, P. M. & Sun, D.-W.
    An accurate water activity model for glycerol solutions and its implementation on moisture sorption isotherm determination.
    Drying Technology 40, 2404-2413 (2022).
    """
    a12 = 0.89924067
    a21 = -1.9173013
    b12 = 127.07511
    b21 = 91.465969
    v1 = 0.0188311
    v2 = 0.08685

    y = 1 - x

    term1 = - np.log(1 + (v2 / v1) * (x / y) * np.exp(-(a12 + b12 / T)))
    term2 = (x / y) / ((v1 / v2) * np.exp(a12 + b12 / T) + x / y)
    term3 = - ((x / y) * (v2 / v1) * np.exp(-(a12 + b12 / T))) / (x / y + (v1 / v2) * np.exp(-(a21 + b21/T)))

    lnaw = term1 + term2 + term3
    return np.exp(lnaw)



"""
Gathers the formulas for the density of glycerol-water solutions.

Sources
-------

- Volk, A. & Kähler, C. J. 
Density model for aqueous glycerol solutions. 
Exp Fluids 59, 75 (2018). 
"""

from ...general import SolutionFormula

from .volk import density_water_glycerol, water_density


class Density_Glycerol_Volk(SolutionFormula):

    source = 'Volk'
    solute = 'glycerol'
    
    temperature_unit = 'C'
    temperature_range = (15, 30)

    concentration_unit = 'w'
    concentration_range = (0, 1)

    with_water_reference = True
    default = True

    def calculate(self, w, T):
        rho_w = water_density(T)
        return rho_w, density_water_glycerol(w, T)


Density_Glycerol_Formulas = (
    Density_Glycerol_Volk,
)


"""Function to calculate the dielectric constant of water.

Sources
-------

--- 'Archer'
    Archer, D. G. & Wang, P.
    The Dielectric Constant of Water and Debye-Hückel Limiting Law Slopes.
    Journal of Physical and Chemical Reference Data 19, 371-411
    (1990)

Also used is this formula for water density, required in Archer:
--- 'IAPWS'
    Patek et al.
    "Reference Correlations for Thermophysical Properties of Liquid Water
    at 0.1 MPa"
    J. Phys. Chem. Ref. Data
    (2009)
"""

import numpy as np
import matplotlib.pyplot as plt
from pynverse import inversefunc

from ...format import make_array_method
from ...constants import Patm
from ..general import WaterFormula
from ..water.density_atm import DensityAtm_IAPWS


rho = DensityAtm_IAPWS()

# I copied all constants from Archer and Wang in case of changes in definitions

# Note: here calculated at Patm, but equation valid for other pressures

alpha = 18.1458392e-30
mu = 6.1375776e-30
Na = 6.0221367e23
k = 1.380658e-23
Mw = 0.0180153
epsilon0 = 8.8541878e-12

b_coeffs = (
    -4.044525e-2,
    103.6180,
    75.32165,
    -23.23778,
    -3.548184,
    - 1246.311,
    263307.7,
    -6.928953e-1,
    -204.4473,
)

T1 = 215
rho_0 = 1000
patm_MPa = Patm / 1e6


def molar_volume(T):
    return Mw / rho.calculate(T=T)


def g_minus_one_over_rho(T, p):
    b1, b2, b3, b4, b5, b6, b7, b8, b9 = b_coeffs
    result = (
        b1 * p / T +
        b2 * T**(-1/2) +
        b3 * (T - T1)**(-1) +
        b4 * (T - T1)**(-1/2) +
        b5 * (T - T1)**(-1/4) +
        np.exp(b6 / T + b7 / T**2 + b8 * p / T + b9 * p / T**2)
    )
    return result


def g(T):
    return 1 + g_minus_one_over_rho(T, p=patm_MPa) * rho.calculate(T=T) / rho_0


def f(T):
    return Na / (3 * molar_volume(T)) * (alpha + g(T) * mu**2 / (3 * epsilon0 * k * T))


def e(epsilon):
    return (epsilon - 1) * (2 * epsilon + 1) / (9 * epsilon)


class DielectricConstant_Archer(WaterFormula):

    source ='IAPWS'
    temperature_unit = 'K'
    # Validity for Archer is muchwider, but the final result is limited by
    # the density formula used (here IAPWS)
    temperature_range = rho.temperature_range
    default = True

    @make_array_method
    def calculate(self, T):
        """Dielectric constant of water according to Archer and Wang (1990)

        Input
        -----
        Temperature in K

        Output
        ------
        Dielectric constant [-]
        """
        f_calc = f(T)
        epsilon_approx = 9 / 2 * f_calc
        epsilon_calc = inversefunc(e, domain=(0.9 * epsilon_approx, 1.1 * epsilon_approx))
        return epsilon_calc(f_calc)


# ========================== WRAP-UP OF FORMULAS =============================

DielectricConstantFormulas = (
    DielectricConstant_Archer,
)



"""Miscellaneous formatting tools and checks for the aquasol package."""

import functools

import numpy as np


def check_units(units, allowed_units):
    """Check if units are among allowed units, raise exception if not."""
    wrong_units = []
    for unit in units:
        if unit not in allowed_units:
            wrong_units.append(unit)
    if len(wrong_units) > 0:
        raise ValueError(f'{wrong_units} not in allowed units {allowed_units}')


def format_input_type(value):
    """If type of input doesn't support math operations, convert to np array."""
    try:
        value + 1
    except TypeError:
        return np.array(value)
    else:
        return value


def format_output_type(value):
    """Formats the output of inverse functions the same way as the inputs.

    This is because e.g. pynverse outputs an array, but sometimes we want it
    to output a scalar if a scalar was put in.
    """
    try:
        sh = value.shape
    except AttributeError:
        return value   # if not an array, it's a float, so return it rightaway
    else:
        if len(sh) == 0:  # this is to return a scalar if a scalar is used as input
            return value.item()
        else:
            return value


def format_temperature(T, unit_in, unit_out):
    """Format temperature from/to Celsius (C) and Kelvin (K)."""

    allowed_units = 'C', 'K'
    check_units([unit_in, unit_out], allowed_units)

    T = format_input_type(T)  # allows for lists and tuples as inputs

    if unit_in == 'C':
        T_celsius = T
        T_kelvin = T + 273.15
    else:
        T_celsius = T - 273.15
        T_kelvin = T

    if unit_out == 'C':
        return T_celsius
    else:
        return T_kelvin


def format_concentration(concentration, unit_out, solute, converter):
    """Check if concentration unit is ok and convert it to the unit_out unit.

    Parameters
    ----------
    concentration: dict from main function **kwargs (e.g. {'w': 0.1})
    unit_out: the unit to format the value into (e.g. 'w')
    solute: name of the solute (e.g. 'NaCl')
    converter: concentration conversion function (convert or basic_convert)

    Output
    ------
    value in the unit_out unit

    Note
    ----
    Checking if concentration is in the right unit etc. and transforming into
    array if input is tuple, list etc. is done by the converter
    """

    if len(concentration) > 1:
        raise ValueError('concentration must have a single keyword argument')

    if len(concentration) == 0:
        raise ValueError(f'Concentration of {solute} not provided.')

    (unit_in, value), = concentration.items()

    conc = converter(value=value, unit1=unit_in, unit2=unit_out, solute=solute)
    return conc


def make_array(function):
    """Decorator to execute function on arrays even when it's not designed initially to accept arrays"""

    @functools.wraps(function)  # to preserve function signature
    def wrapper(x, *args, **kwargs):
        try:
            iter(x)
        except TypeError:  # Not an array
            return function(x, *args, **kwargs)
        else:
            result = []
            for xi in x:
                y = function(xi, *args, **kwargs)
                result.append(y)
            return np.array(result)
    return wrapper


def make_array_method(method):
    """Decorator to execute method on arrays even when it's not designed initially to accept arrays"""

    @functools.wraps(method)  # to preserve method signature
    def wrapper(self, x, *args, **kwargs):
        try:
            iter(x)
        except TypeError:  # Not an array
            return method(self, x, *args, **kwargs)
        else:
            result = []
            for xi in x:
                y = method(self, xi, *args, **kwargs)
                result.append(y)
            return np.array(result)
    return wrapper


def make_array_args(function):
    """Decorator to execute function on arrays even when it's not designed initially to accept arrays.

    This one looks for an iterable in args, not in kwargs.
    Any position of the arg can be iterable, the first one will be chosen if several
    """

    @functools.wraps(function)  # to preserve function signature
    def wrapper(*args, **kwargs):

        for i, val in enumerate(args):
            try:
                iter(val)
            except TypeError: # Not an array --> continue searching
                pass
            else:
                iter_index = i
                iter_val = val
                break
        else:  # No iterable has been found --> apply function directly
            return function(*args, **kwargs)

        result = []

        for val in iter_val:
            new_args = list(args)
            new_args[iter_index] = val
            y = function(*new_args, **kwargs)
            result.append(y)

        return np.array(result)

    return wrapper


def make_array_kwargs(function):
    """Decorator to execute function on arrays even when it's not designed initially to accept arrays

    Here, the array/iterable needs to be passed in the kwargs
    (any kwargs will work, no matter the order)
    If there are several kwargs with iterables, only the first one will be considered
    """

    @functools.wraps(function)  # to preserve function signature
    def wrapper(*args, **kwargs):

        for key, val in kwargs.items():
            try:
                iter(val)
            except TypeError:  # Not an array --> continue searching
                pass
            else:
                iter_key = key
                iter_val = kwargs.pop(key)  # Array: store it elsewhere
                break
        else:  # No iterable has been found --> apply function directly
            return function(*args, **kwargs)

        result = []

        for val in iter_val:
            new_kwargs = {iter_key: val, **kwargs}
            y = function(*args, **new_kwargs)
            result.append(y)

        return np.array(result)

    return wrapper



"""Functions for the calculation of the density of solutions based on Tang 1997

Note:
    Tang 1997 cites other papers (e.g. Tang 1996) of his but compiles data
    in Table I. Some validity ranges (e.g. Na2SO4) are not correct in table I
    and I have taken them from the original sources.

Source
------
- Tang 1997 :
    Tang, I. N.
    Thermodynamic and optical properties of mixed-salt aerosols of atmospheric
    importance. Journal of Geophysical Research 102, 1883-1893 (1997).

Cites :

[NaCl]
Tang, I. N.
Chemical and size effects of hygroscopic aerosols on light scattering
coefficients.
Journal of Geophysical Research: Atmospheres 101, 19245-19250 (1996).

[Na2SO4]
Tang, I. N. & Munkelwitz, H. R.
Simultaneous Determination of Refractive Index and Density of an Evaporating
Aqueous Solution Droplet.
Aerosol Science and Technology 15, 201-207 (1991).

[Also Na2SO4]
Tang, I. N. & Munkelwitz, H. R.
Water activities, densities, and refractive indices of aqueous sulfates and
sodium nitrate droplets of atmospheric importance.
Journal of Geophysical Research 99, 18801 (1994).

[KCl]
Couldn't find the info in these other papers except Tang 1997
"""

from ...general import SolutionFormula


class Density_Tang_Base(SolutionFormula):

    source = 'Tang'
    solute = None  # define in subclasses

    temperature_unit = 'C'
    temperature_range = (25, 25)

    concentration_unit = 'w'
    concentration_range = None  # define in subclasses

    with_water_reference = True

    def calculate(self, w, T):
        """From Tang 1996, only at 25°C"""
        w = w * 100
        rho0 = 997.1  # density of pure water (at 25°C)
        rho = rho0
        for i, coeff in enumerate(self.coeffs):
            rho += coeff * 1000 * w ** (i + 1)
        return rho0, rho


class Density_KCl_Tang_Base(Density_Tang_Base):
    solute = 'KCl'
    concentration_range = (0, 0.44)
    coeffs = (6.13e-3, 4.53e-5, -1.242e-6, 1.582e-8)


class Density_Na2SO4_Tang_Base(Density_Tang_Base):
    solute = 'Na2SO4'
    concentration_range = (0, 0.68)
    coeffs = (8.871e-3, 3.195e-5, 2.28e-7)


class Density_NaCl_Tang_Base(Density_Tang_Base):
    solute = 'NaCl'
    concentration_range = (0, 0.45)
    coeffs = (7.41e-3, -3.741e-5, 2.252e-6, -2.06e-8)
